import datetime as dt
from typing import Generic, TypeVar

import factory
import factory.fuzzy

from django.contrib.auth.models import Group
from django.db.models import Max
from django.utils.timezone import now

from allianceauth.authentication.models import State
from app_utils.testdata_factories import UserMainFactory

from inactivity.models import (
    InactivityPing,
    InactivityPingConfig,
    LeaveOfAbsence,
    Webhook,
)

T = TypeVar("T")


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


# class ResponseFactory(factory.Factory):
#     class Meta:
#         model = requests.Response

#     @factory.post_generation
#     def default_auth_group(self, create, extracted, **kwargs):
#         if not create:
#             return
#         self.status_code = 200
#         self.url = factory.Faker("url")


class GroupFactory(factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Group]):
    class Meta:
        model = Group
        django_get_or_create = ("name",)

    name = factory.Faker("job")

    @factory.post_generation
    def default_auth_group(self, create, extracted, **kwargs):
        if not create:
            return
        self.authgroup.internal = False
        self.authgroup.hidden = False
        self.authgroup.save()


class StateFactory(factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[State]):
    class Meta:
        model = State
        django_get_or_create = ("name", "priority")

    name = factory.Faker("color")

    @factory.lazy_attribute
    def priority(self):
        last_id = State.objects.aggregate(Max("priority"))["priority__max"] or 100
        return last_id + 10

    @factory.post_generation
    def member_characters(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for obj in extracted:
                self.member_characters.add(obj)

    @factory.post_generation
    def member_corporations(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for obj in extracted:
                self.member_corporations.add(obj)

    @factory.post_generation
    def member_alliances(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for obj in extracted:
                self.member_alliances.add(obj)

    @factory.post_generation
    def member_factions(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for obj in extracted:
                self.member_factions.add(obj)


class UserMainRequestorFactory(UserMainFactory):
    """A normal user with access to this app."""

    permissions__ = ["inactivity.basic_access", "memberaudit.basic_access"]

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for group in extracted:
                self.groups.add(group)


class UserMainManagerFactory(UserMainFactory):
    """A user who can manage loa requests."""

    permissions__ = ["inactivity.basic_access", "inactivity.manage_leave"]


class LeaveOfAbsenceFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[LeaveOfAbsence]
):
    class Meta:
        model = LeaveOfAbsence

    class Params:
        is_approved = factory.Trait(
            approver=factory.SubFactory(UserMainManagerFactory), reason=None
        )
        is_denied = factory.Trait(
            approver=factory.SubFactory(UserMainManagerFactory),
            reason=factory.Faker("sentence"),
        )

    user = factory.SubFactory(UserMainRequestorFactory)
    start = factory.fuzzy.FuzzyDateTime(
        start_dt=now() + dt.timedelta(days=1),
        end_dt=now() + dt.timedelta(days=7),
        force_microsecond=0,
    )
    reason = None
    end = factory.LazyAttribute(
        lambda obj: factory.fuzzy.FuzzyDateTime(
            start_dt=obj.start,
            end_dt=obj.start + dt.timedelta(days=7),
            force_microsecond=0,
        ).fuzz()
    )
    notes = factory.Faker("sentence")


class InactivityPingConfigFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[InactivityPingConfig]
):
    class Meta:
        model = InactivityPingConfig

    name = factory.Faker("city")
    days = 3
    text = factory.Faker("sentences")

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for obj in extracted:
                self.groups.add(obj)

    @factory.post_generation
    def states(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for obj in extracted:
                self.states.add(obj)


class InactivityPingFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[InactivityPing]
):
    class Meta:
        model = InactivityPing

    config = factory.SubFactory(InactivityPingConfigFactory)
    user = factory.SubFactory(UserMainRequestorFactory)


class WebhookFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Webhook]
):
    class Meta:
        model = Webhook
        django_get_or_create = ("name", "url")

    name = factory.Faker("color")
    notification_types = [
        Webhook.NotificationType.INACTIVE_USER,
        Webhook.NotificationType.LOA_APPROVED,
        Webhook.NotificationType.LOA_NEW,
    ]

    @factory.lazy_attribute
    def url(self):
        snowflake = factory.fuzzy.FuzzyInteger(
            100000000000000000, 1000000000000000000
        ).fuzz()
        token = factory.fuzzy.FuzzyText(length=68).fuzz()
        url = f"https://discord.com/api/webhooks/{snowflake}/{token}"
        return url

    @factory.post_generation
    def ping_configs(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for obj in extracted:
                self.ping_configs.add(obj)



from django.conf import settings
from django.contrib import admin

from .models import InactivityPing, InactivityPingConfig, LeaveOfAbsence, Webhook


@admin.register(InactivityPingConfig)
class InactivityPingConfigAdmin(admin.ModelAdmin):
    list_display = ("name", "days", "_groups", "_states")
    filter_horizontal = ("groups", "states")

    def _groups(self, obj):
        return ", ".join(obj.groups.values_list("name", flat=True))

    def _states(self, obj):
        return (", ".join(obj.states.values_list("name", flat=True)),)


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    list_display = ("name", "webhook_type", "url")
    filter_horizontal = ("ping_configs",)
    actions = ["send_test_message"]

    @admin.display(description="Send test message to selected webhooks")
    def send_test_message(self, request, queryset):
        for webhook in queryset:
            webhook.send_message("Test message from Inactivity app.")


if settings.DEBUG:

    @admin.register(InactivityPing)
    class InactivityPingAdmin(admin.ModelAdmin):
        list_display = (
            "timestamp",
            "user",
            "config",
        )
        list_filter = (
            ("user", admin.RelatedOnlyFieldListFilter),
            ("config", admin.RelatedOnlyFieldListFilter),
        )
        order = ["-timestamp"]

        def has_add_permission(self, request):
            return False

        def has_change_permission(self, request, obj=None):
            return False

    @admin.register(LeaveOfAbsence)
    class LeaveOfAbsenceAdmin(admin.ModelAdmin):
        list_display = ("pk", "start", "end", "user", "_status", "approver")
        ordering = ("-start",)
        actions = ["approve_leaveofabsence"]

        def get_queryset(self, request):
            qs = super().get_queryset(request)
            return qs.annotate_status().select_related("user", "approver")

        def _status(self, obj) -> bool:
            return obj.status

        @admin.display(description="Approve selected leave of absences")
        def approve_leaveofabsence(self, request, queryset):
            queryset.update(approver=request.user)

        def has_add_permission(self, request):
            return False

        def has_change_permission(self, request, obj=None):
            return False



import datetime as dt
from http import HTTPStatus
from unittest.mock import patch

from django.test import RequestFactory
from django.urls import reverse
from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase, json_response_to_python

from ..views import (
    create_loa_request,
    inactive_users,
    inactive_users_data,
    index,
    list_pending_loa_requests,
    manage_requests,
    my_requests,
)
from .factories import (
    InactivityPingFactory,
    LeaveOfAbsenceFactory,
    UserMainManagerFactory,
    UserMainRequestorFactory,
)

MODULE_PATH = "inactivity.views"


class TestViewsAreWorking(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    def test_index(self):
        # given
        user = UserMainRequestorFactory()
        request = self.factory.get(reverse("inactivity:index"))
        request.user = user
        # when
        response = index(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("inactivity:my_requests"))

    def test_my_request(self):
        # given
        user = UserMainRequestorFactory()
        request = self.factory.get(reverse("inactivity:my_requests"))
        request.user = user
        # when
        response = my_requests(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_managers_can_manage_requests(self):
        # given
        user = UserMainManagerFactory()
        request = self.factory.get(reverse("inactivity:manage_requests"))
        request.user = user
        # when
        response = manage_requests(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_normal_users_can_not_manage_requests(self):
        # given
        user = UserMainRequestorFactory()
        request = self.factory.get(reverse("inactivity:manage_requests"))
        request.user = user
        # when
        response = manage_requests(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)

    def test_list_pending_loa_requests(self):
        # given
        user = UserMainManagerFactory()
        request = self.factory.get(reverse("inactivity:list_pending_loa_requests"))
        request.user = user
        LeaveOfAbsenceFactory(is_approved=True)
        LeaveOfAbsenceFactory(
            start=now() - dt.timedelta(days=2), end=now() - dt.timedelta(days=1)
        )
        unapproved_loa = LeaveOfAbsenceFactory(approver=None, end=None)
        # when
        response = list_pending_loa_requests(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)
        pks = {row["pk"] for row in data["data"]}
        self.assertSetEqual(pks, {unapproved_loa.pk})

    def test_inactive_users(self):
        # given
        InactivityPingFactory()
        user = UserMainManagerFactory()
        request = self.factory.get(reverse("inactivity:inactive_users"))
        request.user = user
        # when
        response = inactive_users(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_inactive_users_data(self):
        # given
        user = UserMainManagerFactory()
        request = self.factory.get(reverse("inactivity:inactive_users_data"))
        request.user = user
        ping = InactivityPingFactory()
        # when
        response = inactive_users_data(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)
        pks = {row["pk"] for row in data["data"]}
        self.assertSetEqual(pks, {ping.pk})


@patch(MODULE_PATH + ".messages")
class TestCreateLoa(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    def test_should_show_empty_form(self, mock_messages):
        # given
        user = UserMainRequestorFactory()
        request = self.factory.get(reverse("inactivity:create_loa_request"))
        request.user = user
        # when
        response = create_loa_request(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_should_create_new_request(self, mock_messages):
        # given
        user = UserMainRequestorFactory()
        request = self.factory.post(
            reverse("inactivity:create_loa_request"),
            data={"start": now().date() + dt.timedelta(days=1), "notes": "dummy"},
        )
        request.user = user
        # when
        response = create_loa_request(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertTrue(mock_messages.info.called)

    def test_should_report_form_errors(self, mock_messages):
        # given
        user = UserMainRequestorFactory()
        request = self.factory.post(
            reverse("inactivity:create_loa_request"),
            data={"start": now().date() - dt.timedelta(days=1)},
        )
        request.user = user
        # when
        response = create_loa_request(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertFalse(mock_messages.info.called)



import humanize
from multiselectfield import MultiSelectField

from django.contrib.auth.models import Group, User
from django.db import models
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import State

from .app_settings import INACTIVITY_TASKS_DEFAULT_PRIORITY
from .helpers import user_for_display
from .managers import InactivityPingConfigManager, LeaveOfAbsenceManager, WebhookManager


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            ("manage_leave", "Can manage leave of absence requests"),
        )


class InactivityPingConfig(models.Model):
    """A ping configuration."""

    name = models.CharField(
        max_length=48,
        unique=True,
        help_text=_("Internal name for the inactivity policy. Must be unique."),
    )
    days = models.PositiveIntegerField(
        help_text=_("The number of days the user must be inactive.")
    )
    text = models.TextField(
        help_text=_("The text of the message or notification sent to the end user.")
    )

    groups = models.ManyToManyField(
        Group,
        blank=True,
        help_text=_(
            "Groups subject to the inactivity policy. If empty, applies to all groups."
        ),
        related_name="+",
    )

    states = models.ManyToManyField(
        State,
        blank=True,
        help_text=_(
            "States subject to the inactivity policy. If empty, applies to all states."
        ),
        related_name="+",
    )

    objects = InactivityPingConfigManager()

    class Meta:
        default_permissions = ()
        verbose_name = _("inactivity policy")
        verbose_name_plural = _("inactivity policies")

    def __str__(self):
        return _("inactivity policy: %(name)s") % {"name": self.name}

    def is_applicable_to(self, user: User) -> bool:
        is_applicable = True
        if self.groups.count() > 0:
            is_applicable &= self.groups.filter(user=user).count() > 0
        if self.states.count() > 0:
            is_applicable &= self.states.filter(userprofile=user.profile).count() > 0
        return is_applicable


class InactivityPing(models.Model):
    """An inactive user who has been notified."""

    config = models.ForeignKey(InactivityPingConfig, on_delete=models.CASCADE)
    last_login_at = models.DateTimeField(null=True, default=None)
    timestamp = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="+")

    class Meta:
        default_permissions = ()

    def __str__(self):
        return _("ping [policy='%(config_name)s' user='%(user_name)s']") % {
            "config_name": self.config.name,
            "user_name": self.user.profile.main_character.character_name
            if self.user.profile.main_character
            else "None",
        }


class LeaveOfAbsence(models.Model):
    """A leave of absence request."""

    class Status(models.TextChoices):
        PENDING = "pending"
        APPROVED = "approved"
        DENIED = "denied"

    approver = models.ForeignKey(
        User, on_delete=models.SET_NULL, blank=True, null=True, related_name="+"
    )
    created_at = models.DateTimeField(null=True, default=None)
    reason = models.TextField(
        blank=True,
        null=True,
        default=None,
        help_text=_("Reason for rejecting a request"),
    )
    end = models.DateField(
        blank=True,
        null=True,
        help_text=_(
            "The end of the leave of absence. Leave blank for an indefinite leave."
        ),
    )
    notes = models.TextField(
        blank=True,
        verbose_name="description",
        help_text=_("Description what this leave of absence request is about."),
    )
    start = models.DateField(help_text=_("The start of the leave of absence."))
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="leave_of_absence_requests",
        help_text="The requestor",
    )

    objects = LeaveOfAbsenceManager()

    class Meta:
        default_permissions = ()

    def __str__(self):
        requestor = user_for_display(self.user)
        return _(
            "%(requestor)s's leave of absence request "
            "from %(start)s to %(end)s %(details)s"
        ) % {
            "end": self.end_text,
            "start": self.start,
            "requestor": requestor.name_with_ticker,
            "details": f'"{self.notes[:25]}"' if self.notes else "",
        }

    def save(self, *args, **kwargs) -> None:
        is_new = self._state.adding is True
        if is_new and not self.created_at:
            self.created_at = now()
        return super().save(*args, **kwargs)

    @property
    def approver_name(self) -> str:
        """Name of the approver."""
        if not self.approver:
            return ""
        approver = user_for_display(self.approver)
        return approver.name_with_ticker

    @property
    def end_text(self) -> str:
        return str(self.end) if self.end else _("open end")

    @property
    def requestor_name(self) -> str:
        """Name of the requestor."""
        requestor = user_for_display(self.user)
        return requestor.name_with_ticker

    def to_output_dict(self) -> dict:
        requestor_display = user_for_display(self.user)
        approver_display = user_for_display(self.approver) if self.approver else None
        duration = (
            humanize.naturaldelta(self.end - self.start) if self.end else _("open end")
        )
        created_at_display = (
            humanize.naturaltime(self.created_at, when=now())
            if self.created_at
            else "?"
        )
        try:
            status = self.status
        except AttributeError:
            status_html = ""
        else:
            if status == self.Status.DENIED:
                status_html = format_html(
                    '<span class="text-danger text-underline-dotted" title="{}">{}</span>',
                    f"Reason: {self.reason}",
                    status,
                )
            else:
                status_html = format_html(
                    '<span class="text-success">{}</span>', status
                )

        result = {
            "approver": approver_display.name if approver_display else "",
            "approver_html": approver_display.html if self.approver else "",
            "created_at": {
                "display": created_at_display,
                "sort": self.created_at.isoformat() if self.created_at else "",
            },
            "duration": duration,
            "end": self.end_text,
            "notes": self.notes if self.notes else "-",
            "requestor": requestor_display.name,
            "requestor_html": requestor_display.html,
            "pk": self.pk,
            "start": self.start,
            "status_html": status_html,
        }
        return result


class Webhook(models.Model):
    "A webhook configuration to send message to."

    class NotificationType(models.IntegerChoices):
        INACTIVE_USER = 1, "Inactive User"
        LOA_NEW = 10, "Leave of Absence - Created"
        LOA_APPROVED = 11, "Leave of Absence - Approved"

    class WebhookType(models.IntegerChoices):
        DISCORD = 1, _("Discord Webhook")

    name = models.CharField(
        max_length=64, unique=True, help_text=_("short name to identify this webhook")
    )

    notification_types = MultiSelectField(
        choices=NotificationType.choices,
        help_text=_(
            "only notifications of the selected types are sent to this webhook"
        ),
    )

    ping_configs = models.ManyToManyField(
        InactivityPingConfig,
        blank=True,
        help_text=_(
            "The inactivity policies to alert for. "
            "If left blank, all policies are alerted for."
        ),
    )

    url = models.CharField(
        max_length=255,
        unique=True,
        help_text=_(
            "URL of this webhook, e.g. "
            "https://discordapp.com/api/webhooks/123456/abcdef"
        ),
    )
    webhook_type = models.IntegerField(
        choices=WebhookType.choices,
        default=WebhookType.DISCORD,
        help_text=_("type of this webhook"),
    )
    is_active = models.BooleanField(
        default=True,
        help_text=_("whether notifications are currently sent to this webhook"),
    )

    objects = WebhookManager()

    def __str__(self):
        return self.name

    def send_message(self, content: str):
        """Send message to webhook asynchronously."""
        from .tasks import send_message_to_webhook

        if self.webhook_type == self.WebhookType.DISCORD:
            send_message_to_webhook.apply_async(
                kwargs={"webhook_pk": self.pk, "content": content},
                priority=INACTIVITY_TASKS_DEFAULT_PRIORITY,
            )



from typing import Any

from django import forms
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from .models import LeaveOfAbsence


class CreateRequestForm(forms.ModelForm):
    def __init__(self, *args, **kwargs) -> None:
        self._user = kwargs.pop("user")
        super().__init__(*args, **kwargs)
        self.fields["notes"].required = True

    class Meta:
        model = LeaveOfAbsence
        fields = ["start", "end", "notes"]

    def clean(self):
        data = super().clean()
        start = data["start"] if "start" in data else None
        end = data["end"] if "end" in data else None
        today = now().date()

        if end and start and end < start:
            raise ValidationError({"end": _("End date must be after the start date.")})

        if start and start < today:
            raise ValidationError({"start": _("Start date can not be in the past.")})

        if start:
            pending = LeaveOfAbsence.objects.filter_pending().filter(user=self._user)
            overlapping = False
            if end:
                if pending.exclude(Q(end__lte=start) | Q(start__gte=end)).exists():
                    overlapping = True
            else:
                if pending.exclude(end__lte=start).exists():
                    overlapping = True

            if overlapping:
                raise ValidationError(
                    _(
                        "There already exists a pending request overlapping with "
                        "this time frame. "
                        "Please adjust the time frame or cancel the existing request."
                    )
                )

    def save(self, *args, **kwargs) -> Any:
        self.instance.user = self._user  # add current user
        return super().save(*args, **kwargs)


class RejectRequestForm(forms.Form):
    request = forms.CharField(
        disabled=True,
        help_text="The leave of absence request to be rejected",
        required=False,
    )
    reason = forms.CharField(
        label="Reason",
        max_length=100,
        required=True,
        help_text="Reason for rejecting this request",
    )



import datetime as dt
from http import HTTPStatus
from unittest.mock import patch

import discord
from celery.exceptions import Retry as CeleryRetry
from memberaudit.tests.testdata.factories import (
    create_character_from_user,
    create_online_status,
)

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from ..models import InactivityPing
from ..tasks import (
    check_inactivity,
    check_inactivity_for_user,
    send_inactivity_ping,
    send_message_to_webhook,
)
from .factories import (
    InactivityPingConfigFactory,
    InactivityPingFactory,
    LeaveOfAbsenceFactory,
    UserMainRequestorFactory,
    WebhookFactory,
)

MODELS_PATH = "inactivity.models"
TASKS_PATH = "inactivity.tasks"


@patch(TASKS_PATH + ".send_message_to_webhook.apply_async", spec=True)
@patch(TASKS_PATH + ".notify.danger", spec=True)
class TestSendInactivityPing(NoSocketsTestCase):
    def test_should_ping_user_and_webhook(
        self, mock_notify_user, mock_send_message_to_webhook
    ):
        # given
        config = InactivityPingConfigFactory()
        user = UserMainRequestorFactory()
        WebhookFactory()
        # when
        send_inactivity_ping(user_pk=user.pk, config_pk=config.pk, last_login_at=now())
        # then
        args, _ = mock_notify_user.call_args
        self.assertEqual(args[0], user)
        self.assertTrue(mock_send_message_to_webhook.called)
        self.assertTrue(
            InactivityPing.objects.filter(
                user=user,
                config=config,
                timestamp__gte=now() - dt.timedelta(seconds=10),
            ).exists()
        )

    def test_should_ping_user_only(
        self, mock_notify_user, mock_send_message_to_webhook
    ):
        # given
        config = InactivityPingConfigFactory()
        user = UserMainRequestorFactory()
        # when
        send_inactivity_ping(user_pk=user.pk, config_pk=config.pk, last_login_at=now())
        # then
        args, _ = mock_notify_user.call_args
        self.assertEqual(args[0], user)
        self.assertFalse(mock_send_message_to_webhook.called)


@patch(TASKS_PATH + ".send_inactivity_ping", spec=True)
class TestCheckInactivityForUser(NoSocketsTestCase):
    def test_should_ping_for_inactive_user(self, mock_send_inactivity_ping):
        # given
        user = UserMainRequestorFactory()
        character = create_character_from_user(user=user)
        last_login = now() - dt.timedelta(days=5)
        create_online_status(
            character=character,
            last_login=last_login,
            last_logout=last_login + dt.timedelta(hours=4),
        )
        InactivityPingConfigFactory(days=3)
        # when
        check_inactivity_for_user(user_pk=user.pk)
        # then
        self.assertTrue(mock_send_inactivity_ping.apply_async.called)

    def test_should_not_ping_for_active_user(self, mock_send_inactivity_ping):
        # given
        user = UserMainRequestorFactory()
        character = create_character_from_user(user=user)
        last_login = now() - dt.timedelta(days=1)
        create_online_status(
            character=character,
            last_login=last_login,
            last_logout=last_login + dt.timedelta(hours=4),
        )
        InactivityPingConfigFactory(days=3)
        # when
        check_inactivity_for_user(user_pk=user.pk)
        # then
        self.assertFalse(mock_send_inactivity_ping.apply_async.called)

    def test_should_not_ping_for_user_without_character(
        self, mock_send_inactivity_ping
    ):
        # given
        user = UserMainRequestorFactory()
        InactivityPingConfigFactory(days=3)
        # when
        check_inactivity_for_user(user_pk=user.pk)
        # then
        self.assertFalse(mock_send_inactivity_ping.apply_async.called)

    def test_should_not_ping_for_excused_user(self, mock_send_inactivity_ping):
        # given
        user = UserMainRequestorFactory()
        character = create_character_from_user(user=user)
        last_login = now() - dt.timedelta(days=4)
        create_online_status(
            character=character,
            last_login=last_login,
            last_logout=last_login + dt.timedelta(hours=4),
        )
        LeaveOfAbsenceFactory(
            user=user,
            start=last_login,
            end=now() + dt.timedelta(days=7),
            is_approved=True,
        )
        InactivityPingConfigFactory(days=3)
        # when
        check_inactivity_for_user(user_pk=user.pk)
        # then
        self.assertFalse(mock_send_inactivity_ping.apply_async.called)

    def test_should_not_ping_when_already_pinged(self, mock_send_inactivity_ping):
        # given
        user = UserMainRequestorFactory()
        character = create_character_from_user(user=user)
        last_login = now() - dt.timedelta(days=4)
        create_online_status(
            character=character,
            last_login=last_login,
            last_logout=last_login + dt.timedelta(hours=4),
        )
        config = InactivityPingConfigFactory(days=3)
        InactivityPingFactory(user=user, config=config)
        # when
        check_inactivity_for_user(user_pk=user.pk)
        # then
        self.assertFalse(mock_send_inactivity_ping.apply_async.called)


@patch(TASKS_PATH + ".check_inactivity_for_user", spec=True)
class TestCheckInactivity(NoSocketsTestCase):
    def test_should_check_inactivity_for_registered_users_only(
        self, mock_check_inactivity_for_user
    ):
        # given
        InactivityPingConfigFactory()
        user = UserMainRequestorFactory()
        create_character_from_user(user=user)
        UserMainRequestorFactory()  # will not be checked
        # when
        check_inactivity()
        # then
        users_pks_checked = {
            obj[1]["kwargs"]["user_pk"]
            for obj in mock_check_inactivity_for_user.apply_async.call_args_list
        }
        self.assertSetEqual(users_pks_checked, {user.pk})

    def test_should_do_nothing_when_no_config(self, mock_check_inactivity_for_user):
        # given
        user = UserMainRequestorFactory()
        create_character_from_user(user=user)
        UserMainRequestorFactory()  # will not be checked
        # when
        check_inactivity()
        # then
        users_pks_checked = {
            obj[1]["kwargs"]["user_pk"]
            for obj in mock_check_inactivity_for_user.apply_async.call_args_list
        }
        self.assertSetEqual(users_pks_checked, set())


class FakeResponse:
    def __init__(self, status: int, headers: dict = None):
        self.status = int(status)
        self.reason = "Dummy reason"
        self.text = "Dummy text"
        if headers:
            self.headers = {str(key): str(value) for key, value in headers.items()}


@patch(TASKS_PATH + ".cache.lock", spec=True)
@patch(TASKS_PATH + ".discord.SyncWebhook.send", spec=True)
class TestSendMessageToWebhook(NoSocketsTestCase):
    def test_send_message(self, mock_send, mock_cache_lock):
        # given
        webhook = WebhookFactory()
        # when
        send_message_to_webhook(webhook.pk, "dummy")
        # then
        self.assertTrue(mock_send.called)

    def test_retry_when_rate_limited(self, mock_send, mock_cache_lock):
        # given
        response = FakeResponse(
            status=HTTPStatus.TOO_MANY_REQUESTS,
            headers={"Retry-After": 99},
        )
        my_exception = discord.HTTPException(
            response=response, message="Test exception"
        )
        mock_send.side_effect = my_exception
        webhook = WebhookFactory()
        # when/then
        with self.assertRaises(CeleryRetry):
            send_message_to_webhook(webhook.pk, "dummy")

    def test_raise_error_when_other_http_error(self, mock_send, mock_cache_lock):
        # given
        response = FakeResponse(status=HTTPStatus.INTERNAL_SERVER_ERROR)
        my_exception = discord.HTTPException(
            response=response, message="Test exception"
        )
        mock_send.side_effect = my_exception
        webhook = WebhookFactory()
        # when/then
        with self.assertRaises(discord.HTTPException):
            send_message_to_webhook(webhook.pk, "dummy")



from app_utils.testing import NoSocketsTestCase

from ..models import InactivityPingConfig, LeaveOfAbsence, Webhook
from .factories import (
    GroupFactory,
    InactivityPingConfigFactory,
    LeaveOfAbsenceFactory,
    UserMainRequestorFactory,
    WebhookFactory,
)


class TestLeaveOfAbsenceManager(NoSocketsTestCase):
    def test_should_ping_for_inactive_user(self):
        # given
        LeaveOfAbsenceFactory(is_approved=True)
        LeaveOfAbsenceFactory()
        # when
        self.assertEqual(LeaveOfAbsence.objects.unapproved_count(), 1)

    def test_should_annotate_status_correctly(self):
        # given
        pending = LeaveOfAbsenceFactory()
        approved = LeaveOfAbsenceFactory(is_approved=True)
        denied = LeaveOfAbsenceFactory(is_denied=True)
        # when
        result = {obj.pk: obj for obj in LeaveOfAbsence.objects.annotate_status()}
        # then
        self.assertEqual(result[pending.pk].status, LeaveOfAbsence.Status.PENDING)
        self.assertEqual(result[approved.pk].status, LeaveOfAbsence.Status.APPROVED)
        self.assertEqual(result[denied.pk].status, LeaveOfAbsence.Status.DENIED)


class TestInactivityPingConfigManager(NoSocketsTestCase):
    def test_should_filter_relevant_for_user(self):
        # given
        config_1 = InactivityPingConfigFactory()
        group_1 = GroupFactory()
        config_2 = InactivityPingConfigFactory(groups=[group_1])
        group_2 = GroupFactory()
        InactivityPingConfigFactory(groups=[group_2])

        user = UserMainRequestorFactory(groups=[group_1])
        # when
        result = InactivityPingConfig.objects.relevant_for_user(user)
        # then
        result_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(result_pks, {config_1.pk, config_2.pk})


class TestWebhookManager(NoSocketsTestCase):
    def test_should_filter_relevant_for_user(self):
        # given
        group_1 = GroupFactory()
        config_1 = InactivityPingConfigFactory(groups=[group_1])
        webhook = WebhookFactory(ping_configs=[config_1])
        group_2 = GroupFactory()
        config_2 = InactivityPingConfigFactory(groups=[group_2])
        WebhookFactory(ping_configs=[config_2])
        user = UserMainRequestorFactory(groups=[group_1])
        # when
        result = Webhook.objects.relevant_for_user(user)
        # then
        result_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(result_pks, {webhook.pk})



import datetime as dt
from http import HTTPStatus

import discord
import humanize
import requests
from celery import Task, shared_task
from memberaudit.models import Character

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db.models import Max, Q
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import INACTIVITY_TASKS_DEFAULT_PRIORITY
from .models import InactivityPing, InactivityPingConfig, Webhook

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class ModifiedSession(requests.Session):
    """Modified requests session for adding a timeout
    when sending message to a Webhook.
    """

    def post(self, *args, **kwargs) -> requests.Response:
        return super().post(*args, timeout=30, **kwargs)


@shared_task
def check_inactivity():
    """Check all users with registered Member Audit characters for inactivity.
    This is the main periodic task.
    """
    if not InactivityPingConfig.objects.exists():
        logger.warning("No inactivity config found. Nothing to do.")
        return

    registered_users = User.objects.filter(
        character_ownerships__character__memberaudit_character__isnull=False
    ).distinct()
    for user_pk in registered_users.values_list("pk", flat=True):
        check_inactivity_for_user.apply_async(
            kwargs={"user_pk": user_pk}, priority=INACTIVITY_TASKS_DEFAULT_PRIORITY
        )


@shared_task
def check_inactivity_for_user(user_pk: int):
    today = now().date()
    user = User.objects.get(pk=user_pk)
    if not (
        user.leave_of_absence_requests.filter(
            Q(start__lt=today), Q(end=None) | Q(end__gt=today), ~Q(approver=None)
        ).exists()
    ):
        last_loa = (
            user.leave_of_absence_requests.filter(Q(end__lt=today), ~Q(approver=None))
            .order_by("-end")
            .first()
        )
        for config in InactivityPingConfig.objects.relevant_for_user(user):
            threshold_date = today - dt.timedelta(days=config.days)
            characters = Character.objects.owned_by_user(user)
            registered = characters.exists()
            active = characters.filter(
                Q(online_status__last_login__gt=threshold_date)
                | Q(online_status__last_logout__gt=threshold_date),
            ).exists()
            excused = last_loa and (not last_loa.end or threshold_date < last_loa.end)
            pinged = InactivityPing.objects.filter(
                user__pk=user_pk, config=config
            ).exists()
            if active:
                InactivityPing.objects.filter(user__pk=user_pk, config=config).delete()
            if not active and registered and not pinged and not excused:
                last_login_at = characters.aggregate(
                    Max("online_status__last_login")
                ).get("online_status__last_login__max")
                send_inactivity_ping.apply_async(
                    kwargs={
                        "user_pk": user_pk,
                        "config_pk": config.pk,
                        "last_login_at": last_login_at,
                    },
                    priority=INACTIVITY_TASKS_DEFAULT_PRIORITY,
                )


@shared_task
def send_inactivity_ping(user_pk: int, config_pk: int, last_login_at: dt.datetime):
    config = InactivityPingConfig.objects.get(pk=config_pk)
    user = User.objects.get(pk=user_pk)
    notify.danger(user, title="Inactivity notification", message=config.text)
    InactivityPing.objects.create(config=config, user=user, last_login_at=last_login_at)
    for webhook in Webhook.objects.filter(
        Q(ping_configs=config) | Q(ping_configs=None),
        Q(is_active=True),
        notification_types__contains=Webhook.NotificationType.INACTIVE_USER,
    ):
        duration = humanize.naturaldelta(now() - last_login_at)
        message = _(
            "**%(user_name)s** has been inactive for **%(duration)s** "
            "and has been notified according to **%(config_name)s** policy"
        ) % {
            "user_name": user.profile.main_character.character_name,
            "duration": duration,
            "config_name": config.name,
        }
        send_message_to_webhook.apply_async(
            kwargs={"webhook_pk": webhook.pk, "content": message},
            priority=INACTIVITY_TASKS_DEFAULT_PRIORITY,
        )


@shared_task(bind=True, autoretry_for=(requests.Timeout,), retry_backoff=True)
def send_message_to_webhook(self: Task, webhook_pk: int, content: str):
    """Send a message to the webhook"""
    webhook = Webhook.objects.get(pk=webhook_pk)
    logger.info("Sending message to webhook: %s", webhook)
    with ModifiedSession() as session:
        hook = discord.SyncWebhook.from_url(webhook.url, session=session)
        try:
            with cache.lock(f"inactivity-lock-webhook-{webhook.pk}"):
                hook.send(content=content)
        except discord.HTTPException as ex:
            if ex.status == HTTPStatus.TOO_MANY_REQUESTS:
                try:
                    retry_after = int(ex.response.headers["Retry-After"])
                except KeyError:
                    retry_after = 60
                logger.error(
                    "%s: Rate limited. Trying again in %s seconds. Error: %s",
                    webhook,
                    retry_after,
                    ex.response.text,
                )
                raise self.retry(countdown=retry_after) from ex
            raise ex



from typing import Optional

import humanize

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from allianceauth.notifications import notify

from .forms import CreateRequestForm, RejectRequestForm
from .helpers import user_for_display
from .models import InactivityPing, InactivityPingConfig, LeaveOfAbsence, Webhook


def add_common_context(request, context: Optional[dict] = None) -> dict:
    """adds the common context used by all view"""
    context = context or {}
    new_context = {
        **{
            "app_title": "Leave of Absence",
            "unapproved_count": LeaveOfAbsence.objects.unapproved_count(),
        },
        **context,
    }
    return new_context


def make_action_button_html(
    label: str,
    url: str,
    button_type: str = "default",
    tooltip: str = "",
    disabled: bool = False,
) -> str:
    """Make HTML for an action button."""
    return format_html(
        '<a href="{}" class="btn btn-{}"{}{}>{}</a>',
        url,
        button_type,
        format_html(' title="{}"', tooltip) if tooltip else "",
        format_html(' disabled="disabled"') if disabled else "",
        label,
    )


@login_required
@permission_required("inactivity.basic_access")
def index(request):
    return redirect("inactivity:my_requests")


# ---- my requests -----


@login_required
@permission_required("inactivity.basic_access")
def my_requests(request):
    context = {"page_title": _("My Requests")}
    return render(
        request, "inactivity/my_requests.html", add_common_context(request, context)
    )


@login_required
@permission_required("inactivity.basic_access")
def my_open_requests_data(request):
    data = []
    for loa_request in LeaveOfAbsence.objects.filter_pending().filter(
        user=request.user
    ):
        row = loa_request.to_output_dict()
        actions = make_action_button_html(
            label=_("Cancel"),
            url=reverse("inactivity:cancel_loa_request", args=[loa_request.pk]),
            button_type="danger",
            tooltip="Cancel this request",
        )
        row["actions"] = actions
        data.append(row)
    return JsonResponse({"data": data})


@login_required
@permission_required("inactivity.basic_access")
def my_completed_requests_data(request):
    data = []
    for loa_request in (
        LeaveOfAbsence.objects.filter_processed()
        .filter(user=request.user)
        .annotate_status()
    ):
        row = loa_request.to_output_dict()
        data.append(row)
    return JsonResponse({"data": data})


@login_required
@permission_required("inactivity.basic_access")
def create_loa_request(request):
    if request.method == "POST":
        form = CreateRequestForm(request.POST, user=request.user)
        if form.is_valid():
            loa: LeaveOfAbsence = form.save()
            messages.info(
                request,
                format_html(_("Your request has been submitted for review: %s") % loa),
            )
            message = _("New request: %s") % loa
            Webhook.objects.send_message_to_active_webhooks(
                loa, Webhook.NotificationType.LOA_NEW, message
            )
            return redirect("inactivity:index")

    else:
        form = CreateRequestForm(user=request.user)

    context = {"form": form}
    return render(request, "inactivity/create_loa.html", context)


@login_required
@permission_required("inactivity.basic_access")
def cancel_loa_request(request, loa_request_pk):
    loa = get_object_or_404(LeaveOfAbsence, pk=loa_request_pk, user=request.user)
    loa.delete()
    messages.info(request, format_html(_("%s has been canceled.") % loa))
    return redirect("inactivity:index")


# ---- manage requests -----


@login_required
@permission_required("inactivity.manage_leave")
def manage_requests(request):
    context = {"page_title": _("Manage Requests")}
    return render(
        request, "inactivity/manage_requests.html", add_common_context(request, context)
    )


@login_required
@permission_required("inactivity.manage_leave")
def list_pending_loa_requests(request):
    data = []
    for loa_request in LeaveOfAbsence.objects.filter_pending():
        row = loa_request.to_output_dict()
        actions = [
            make_action_button_html(
                label="Approve",
                url=reverse("inactivity:approve_loa_request", args=[loa_request.pk]),
                button_type="success",
                tooltip=_("Approve this request"),
            ),
            make_action_button_html(
                label=_("Deny"),
                url=reverse("inactivity:deny_loa_request", args=[loa_request.pk]),
                button_type="danger",
                tooltip=_("Deny this request"),
            ),
        ]
        row["actions"] = " ".join(actions)
        data.append(row)
    return JsonResponse({"data": data})


@login_required
@permission_required("inactivity.manage_leave")
def list_processed_loa_requests(request):
    data = []
    for loa_request in LeaveOfAbsence.objects.filter_processed().annotate_status():
        row = loa_request.to_output_dict()
        row["actions"] = ""
        data.append(row)
    return JsonResponse({"data": data})


@login_required
@permission_required("inactivity.manage_leave")
def approve_loa_request(request, loa_request_pk):
    loa: LeaveOfAbsence = get_object_or_404(LeaveOfAbsence, pk=loa_request_pk)
    loa.approver = request.user
    loa.save()
    messages.success(request, format_html(_("Your have approved: %s") % loa))
    message = _("%(loa)s has been approved by %(approver_name)s") % {
        "loa": loa,
        "approver_name": loa.approver_name,
    }
    notify.success(loa.user, title="LOA approved", message=message)
    Webhook.objects.send_message_to_active_webhooks(
        loa, Webhook.NotificationType.LOA_APPROVED, message
    )
    return redirect("inactivity:manage_requests")


@login_required
@permission_required("inactivity.manage_leave")
def deny_loa_request(request, loa_request_pk):
    loa_request: LeaveOfAbsence = get_object_or_404(LeaveOfAbsence, pk=loa_request_pk)
    requestor = loa_request.user

    if request.method == "POST":
        form = RejectRequestForm(request.POST)
        if form.is_valid():
            loa_request.approver = request.user
            loa_request.reason = form.cleaned_data["reason"]
            loa_request.save()
            messages.info(request, _("%s has been denied") % loa_request)
            manager_display = user_for_display(loa_request.approver)
            message = _(
                "%(loa)s has been denied by %(manager)s. Reason: %(reason)s"
            ) % {
                "loa": loa_request,
                "manager": manager_display.name,
                "reason": loa_request.reason,
            }
            notify.danger(requestor, "LOA rejected", message)
            return redirect("inactivity:manage_requests")

    else:
        form = RejectRequestForm(initial={"request": str(loa_request)})

    return render(
        request,
        "inactivity/reject_request.html",
        {"form": form, "loa_request": loa_request},
    )


# ---- inactive users -----


@login_required
@permission_required("inactivity.basic_access")
def inactive_users(request):
    has_policies = InactivityPingConfig.objects.exists()
    context = {"page_title": _("Inactive users"), "has_policies": has_policies}
    return render(
        request, "inactivity/inactive_users.html", add_common_context(request, context)
    )


@login_required
@permission_required("inactivity.basic_access")
def inactive_users_data(request):
    data = []
    for obj in InactivityPing.objects.all():
        obj: InactivityPing
        user_obj = user_for_display(obj.user)
        row = {
            "pk": obj.pk,
            "user_html": {"display": user_obj.html, "sort": user_obj.name},
            "last_login_at": {
                "display": humanize.naturaltime(obj.last_login_at, when=now())
                if obj.last_login_at
                else "?",
                "sort": obj.last_login_at.isoformat() if obj.last_login_at else None,
            },
            "policy": obj.config.name,
            "notified_at": {
                "display": humanize.naturaltime(obj.timestamp, when=now()),
                "sort": obj.timestamp.isoformat(),
            },
        }
        data.append(row)
    return JsonResponse({"data": data})



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from .factories import (
    GroupFactory,
    InactivityPingConfigFactory,
    InactivityPingFactory,
    LeaveOfAbsenceFactory,
    StateFactory,
    UserMainRequestorFactory,
    WebhookFactory,
)

MODELS_PATH = "inactivity.models"
TASKS_PATH = "inactivity.tasks"


class TestInactivityPing(NoSocketsTestCase):
    def test_str_normal(self):
        # given
        ping = InactivityPingFactory()
        # when
        result = str(ping)
        # then
        self.assertIn(ping.user.last_name, result)

    def test_str_user_without_main(self):
        # given
        ping = InactivityPingFactory(user=UserMainRequestorFactory())
        # when
        result = str(ping)
        # then
        self.assertTrue(result)


class TestInactivityPingConfig(NoSocketsTestCase):
    def test_str_normal(self):
        # given
        ping = InactivityPingConfigFactory()
        # when
        result = str(ping)
        # then
        self.assertTrue(result)


class TestInactivityPingConfigIsApplicableTo(NoSocketsTestCase):
    def test_true_when_no_group_or_state_set(self):
        # given
        config = InactivityPingConfigFactory()
        user = UserMainRequestorFactory()
        # when/then
        self.assertTrue(config.is_applicable_to(user))

    def test_true_when_group_is_matching(self):
        # given
        group = GroupFactory()
        config = InactivityPingConfigFactory(groups=[group])
        user = UserMainRequestorFactory(groups=[group])
        # when/then
        self.assertTrue(config.is_applicable_to(user))

    def test_false_when_group_is_not_matching(self):
        # given
        group_1 = GroupFactory()
        group_2 = GroupFactory()
        config = InactivityPingConfigFactory(groups=[group_1])
        user = UserMainRequestorFactory(groups=[group_2])
        # when/then
        self.assertFalse(config.is_applicable_to(user))

    def test_true_when_state_is_matching(self):
        # given
        user = UserMainRequestorFactory()
        state = StateFactory(member_characters=[user.profile.main_character])
        config = InactivityPingConfigFactory(states=[state])
        # when/then
        self.assertTrue(config.is_applicable_to(user))

    def test_false_when_state_is_not_matching(self):
        # given
        user = UserMainRequestorFactory()
        StateFactory(member_characters=[user.profile.main_character])
        state_2 = StateFactory()
        config = InactivityPingConfigFactory(states=[state_2])
        # when/then
        self.assertFalse(config.is_applicable_to(user))


class TestLeaveOfAbsence(NoSocketsTestCase):
    def test_str(self):
        # given
        loa = LeaveOfAbsenceFactory()
        # when/then
        self.assertIsInstance(str(loa), str)

    def test_should_set_created_at_when_created(self):
        # when
        my_now = now()
        with patch(MODELS_PATH + ".now", lambda: my_now):
            obj = LeaveOfAbsenceFactory()
        # then
        self.assertEqual(obj.created_at, my_now)

    def test_should_not_update_created_at_when_changed(self):
        # given
        my_dt = now() - dt.timedelta(days=3)
        obj = LeaveOfAbsenceFactory(created_at=my_dt)
        # when
        obj.save()
        # then
        self.assertEqual(obj.created_at, my_dt)


@patch(TASKS_PATH + ".send_message_to_webhook", spec=True)
class TestWebhook(NoSocketsTestCase):
    def test_send_message(self, mock_send_message_to_webhook):
        # given
        webhook = WebhookFactory()
        # when
        webhook.send_message("dummy")
        # then
        self.assertTrue(mock_send_message_to_webhook.apply_async.called)



from typing import NamedTuple, Optional

from allianceauth.authentication.models import User
from allianceauth.eveonline.models import EveCharacter
from app_utils.views import bootstrap_icon_plus_name_html


class UserForDisplay(NamedTuple):
    """User with additional information for display."""

    character: Optional[EveCharacter]
    html: str
    icon_url: str
    name: str
    name_with_ticker: str
    user: User

    def has_main(self) -> bool:
        return bool(self.character)


def main_or_username(user: User) -> str:
    """Safe way to get the main's character name or fallback to the username."""
    try:
        return user.profile.main_character.character_name
    except AttributeError:
        return user.username


def user_for_display(user: User) -> UserForDisplay:
    """Prepare user for display.

    Use main character as default or fallback to username if user has no main.
    """
    character: EveCharacter = user.profile.main_character
    if not character:
        name = name_with_ticker = user.username
        icon_url = EveCharacter.generic_portrait_url(1)  # empty portrait
    else:
        name = character.character_name
        name_with_ticker = f"{name} [{character.corporation_ticker}]"
        icon_url = character.portrait_url()
    html = bootstrap_icon_plus_name_html(icon_url, name_with_ticker, avatar=True)
    return UserForDisplay(
        name_with_ticker=name_with_ticker,
        name=name,
        html=html,
        character=character,
        icon_url=icon_url,
        user=user,
    )



from django.contrib.auth.models import User
from django.db import models
from django.db.models import Case, Q, Value, When
from django.utils.timezone import now


class LeaveOfAbsenceQuerySet(models.QuerySet):
    def filter_pending(self):
        """Filter all pending requests."""
        return self.filter(Q(end=None) | Q(end__gt=now()), Q(approver=None))

    def filter_processed(self):
        """Filter all processed requests."""
        return self.filter(approver__isnull=False)

    def annotate_status(self):
        Status = self.model.Status
        return self.annotate(
            status=Case(
                When(approver__isnull=True, then=Value(Status.PENDING)),
                When(reason__isnull=False, then=Value(Status.DENIED)),
                default=Value(Status.APPROVED),
            )
        )


class LeaveOfAbsenceManagerBase(models.Manager):
    def unapproved_count(self):
        """Count number of unapproved requests."""
        return self.filter_pending().count()


LeaveOfAbsenceManager = LeaveOfAbsenceManagerBase.from_queryset(LeaveOfAbsenceQuerySet)


class InactivityPingConfigQueryset(models.QuerySet):
    def relevant_for_user(self, user: User):
        """Webhooks that are relevant for the given user."""
        pks = {config.pk for config in self.all() if config.is_applicable_to(user)}
        return self.filter(pk__in=pks)


class InactivityPingConfigManagerBase(models.Manager):
    pass


InactivityPingConfigManager = InactivityPingConfigManagerBase.from_queryset(
    InactivityPingConfigQueryset
)


class WebhookQueryset(models.QuerySet):
    def relevant_for_user(self, user: User):
        """Webhooks that are relevant for the given user."""
        from .models import InactivityPingConfig

        configs = list(InactivityPingConfig.objects.relevant_for_user(user))
        return self.filter(Q(ping_configs__in=configs) | Q(ping_configs=None))


class WebhookManagerBase(models.Manager):
    def send_message_to_active_webhooks(self, loa, notif_type, message):
        webhooks = self.relevant_for_user(loa.user).filter(
            is_active=True, notification_types__contains=notif_type
        )
        for webhook in webhooks:
            webhook.send_message(message)


WebhookManager = WebhookManagerBase.from_queryset(WebhookQueryset)



import datetime as dt
from http import HTTPStatus
from unittest.mock import patch

from memberaudit.tests.testdata.factories import (
    create_character_from_user,
    create_online_status,
)

from django.test import override_settings
from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from ..models import LeaveOfAbsence, Webhook
from ..tasks import check_inactivity
from .factories import (
    GroupFactory,
    InactivityPingConfigFactory,
    LeaveOfAbsenceFactory,
    UserMainManagerFactory,
    UserMainRequestorFactory,
    WebhookFactory,
)

MODELS_PATH = "inactivity.models"
TASKS_PATH = "inactivity.tasks"
VIEWS_PATH = "inactivity.views"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(TASKS_PATH + ".discord.SyncWebhook.send", spec=True)
@patch(VIEWS_PATH + ".messages", spec=True)
class TestFrontend(NoSocketsTestCase):
    def test_should_create_new_request_and_notify_when_matching_webhook(
        self, mock_messages, mock_send
    ):
        # given
        group = GroupFactory()
        config = InactivityPingConfigFactory(groups=[group])
        WebhookFactory(
            ping_configs=[config], notification_types=[Webhook.NotificationType.LOA_NEW]
        )
        user = UserMainRequestorFactory(groups=[group])
        self.client.force_login(user)
        start = now().date() + dt.timedelta(days=3)
        end = start + dt.timedelta(days=7)
        # when
        response = self.client.post(
            "/inactivity/loa_requests/create",
            data={"start": start, "end": end, "notes": "test"},
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, "/inactivity/")
        obj = LeaveOfAbsence.objects.first()
        self.assertEqual(obj.user, user)
        self.assertEqual(obj.start, start)
        self.assertEqual(obj.end, end)
        self.assertEqual(obj.notes, "test")
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_send.called)

    def test_should_create_new_request_and_not_notify_when_no_matching_webhook(
        self, mock_messages, mock_send
    ):
        # given
        group = GroupFactory()
        config = InactivityPingConfigFactory(groups=[group])
        WebhookFactory(
            ping_configs=[config],
            notification_types=[Webhook.NotificationType.LOA_APPROVED],
        )
        user = UserMainRequestorFactory()
        self.client.force_login(user)
        start = now().date() + dt.timedelta(days=3)
        # when
        self.client.post(
            "/inactivity/loa_requests/create", data={"start": start, "notes": "test"}
        )
        # then
        self.assertTrue(LeaveOfAbsence.objects.filter(user=user, start=start).exists())
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_send.called)

    def test_should_approve_request_and_notify(self, mock_messages, mock_send):
        # given
        WebhookFactory()
        loa = LeaveOfAbsenceFactory()
        user = UserMainManagerFactory()
        self.client.force_login(user)
        # when
        response = self.client.get(f"/inactivity/loa_requests/{loa.pk}/approve")
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, "/inactivity/manage_requests")
        loa.refresh_from_db()
        self.assertEqual(loa.approver, user)
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_send)

    def test_manager_can_reject_a_request(self, mock_messages, mock_send):
        # given
        WebhookFactory()
        loa = LeaveOfAbsenceFactory()
        manager = UserMainManagerFactory()
        self.client.force_login(manager)
        # when
        response = self.client.post(
            f"/inactivity/loa_requests/{loa.pk}/reject",
            data={"reason": "do not like you much"},
        )
        # then
        loa.refresh_from_db()
        self.assertEqual(loa.approver, manager)
        self.assertTrue(loa.reason)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, "/inactivity/manage_requests")


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(TASKS_PATH + ".discord.SyncWebhook.send", spec=True)
@patch(TASKS_PATH + ".notify.danger", spec=True)
class TestTasksEnd2End(NoSocketsTestCase):
    def test_check_inactivity_e2e(self, mock_notify_user, mock_send):
        # given
        user = UserMainRequestorFactory()
        character = create_character_from_user(user=user)
        last_login = now() - dt.timedelta(days=5)
        create_online_status(
            character=character,
            last_login=last_login,
            last_logout=last_login + dt.timedelta(hours=4),
        )
        InactivityPingConfigFactory(days=3)
        WebhookFactory()
        mock_notify_user.reset_mock()
        mock_send.reset_mock()
        # when
        check_inactivity.delay()
        # then
        self.assertTrue(mock_notify_user.called)
        self.assertTrue(mock_send.called)



import datetime as dt

from django.test import TestCase
from django.utils.timezone import now

from inactivity.forms import CreateRequestForm

from .factories import LeaveOfAbsenceFactory, UserMainRequestorFactory


class TestForms(TestCase):
    def test_should_create_minimal_form(self):
        # given
        form_data = {"start": now() + dt.timedelta(days=1), "notes": "dummy"}
        user = UserMainRequestorFactory()
        # when
        form = CreateRequestForm(data=form_data, user=user)
        # then
        self.assertTrue(form.is_valid())

    def test_should_not_allow_empty_form(self):
        # given
        form_data = {}
        user = UserMainRequestorFactory()
        # when
        form = CreateRequestForm(data=form_data, user=user)
        # then
        self.assertFalse(form.is_valid())

    def test_should_not_allow_end_before_start(self):
        # given
        form_data = {
            "start": now() + dt.timedelta(days=3),
            "end": now() + dt.timedelta(days=1),
            "notes": "dummy",
        }
        user = UserMainRequestorFactory()
        # when
        form = CreateRequestForm(data=form_data, user=user)
        # then
        self.assertFalse(form.is_valid())

    def test_should_not_allow_start_date_in_the_past(self):
        # given
        user = UserMainRequestorFactory()
        form_data = {
            "start": now() - dt.timedelta(days=3),
            "end": now() + dt.timedelta(days=1),
            "notes": "dummy",
        }
        # when
        form = CreateRequestForm(data=form_data, user=user)
        # then
        self.assertFalse(form.is_valid())


class TestFormsOverlapping(TestCase):
    def setUp(self) -> None:
        self.today = now().date()

    def _some_date(self, delta_days: int):
        return self.today + dt.timedelta(days=delta_days)

    def _create_form(self, start: int, end: int):
        return {
            "start": self._some_date(start),
            "end": self._some_date(end),
            "notes": "dummy",
        }

    def test_should_not_allow_inside_existing(self):
        # given
        loa_request = LeaveOfAbsenceFactory(
            start=self._some_date(1), end=self._some_date(7)
        )
        form_data = self._create_form(2, 3)
        # when
        form = CreateRequestForm(data=form_data, user=loa_request.user)
        # then
        self.assertFalse(form.is_valid())

    def test_should_not_allow_overlap_at_beginning(self):
        # given
        loa_request = LeaveOfAbsenceFactory(
            start=self._some_date(1), end=self._some_date(4)
        )
        form_data = self._create_form(3, 7)
        # when
        form = CreateRequestForm(data=form_data, user=loa_request.user)
        # then
        self.assertFalse(form.is_valid())

    def test_should_not_allow_overlap_at_end(self):
        # given
        loa_request = LeaveOfAbsenceFactory(
            start=self._some_date(3), end=self._some_date(7)
        )
        form_data = self._create_form(1, 4)
        # when
        form = CreateRequestForm(data=form_data, user=loa_request.user)
        # then
        self.assertFalse(form.is_valid())

    def test_should_not_allow_encompassing_existing(self):
        # given
        loa_request = LeaveOfAbsenceFactory(
            start=self._some_date(3), end=self._some_date(4)
        )
        form_data = self._create_form(1, 7)
        # when
        form = CreateRequestForm(data=form_data, user=loa_request.user)
        # then
        self.assertFalse(form.is_valid())

    def test_should_allow_existing_before(self):
        # given
        loa_request = LeaveOfAbsenceFactory(
            start=self._some_date(3), end=self._some_date(4)
        )
        form_data = self._create_form(4, 7)
        # when
        form = CreateRequestForm(data=form_data, user=loa_request.user)
        # then
        self.assertTrue(form.is_valid())

    def test_should_allow_existing_after(self):
        # given
        loa_request = LeaveOfAbsenceFactory(
            start=self._some_date(4), end=self._some_date(7)
        )
        form_data = self._create_form(1, 4)
        # when
        form = CreateRequestForm(data=form_data, user=loa_request.user)
        # then
        self.assertTrue(form.is_valid())

    def test_should_ignore_overlaps_from_other_users(self):
        # given
        current_user = UserMainRequestorFactory()
        LeaveOfAbsenceFactory(start=self._some_date(3), end=self._some_date(5))
        form_data = self._create_form(3, 7)
        # when
        form = CreateRequestForm(data=form_data, user=current_user)
        # then
        self.assertTrue(form.is_valid())



from django.urls import path

from . import views

app_name = "inactivity"

urlpatterns = [
    path("", views.index, name="index"),
    path("my_requests", views.my_requests, name="my_requests"),
    path("manage_requests", views.manage_requests, name="manage_requests"),
    path(
        "open_requests_data", views.my_open_requests_data, name="my_open_requests_data"
    ),
    path(
        "competed_requests_data",
        views.my_completed_requests_data,
        name="my_completed_requests_data",
    ),
    path(
        "loa_requests/pending",
        views.list_pending_loa_requests,
        name="list_pending_loa_requests",
    ),
    path(
        "loa_requests/approved",
        views.list_processed_loa_requests,
        name="list_processed_loa_requests",
    ),
    path("loa_requests/create", views.create_loa_request, name="create_loa_request"),
    path(
        "loa_requests/<int:loa_request_pk>/cancel",
        views.cancel_loa_request,
        name="cancel_loa_request",
    ),
    path(
        "loa_requests/<int:loa_request_pk>/approve",
        views.approve_loa_request,
        name="approve_loa_request",
    ),
    path(
        "loa_requests/<int:loa_request_pk>/reject",
        views.deny_loa_request,
        name="deny_loa_request",
    ),
    path("inactive_users", views.inactive_users, name="inactive_users"),
    path("inactive_users_data", views.inactive_users_data, name="inactive_users_data"),
]



# Generated by Django 4.0.10 on 2023-06-12 12:06

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("inactivity", "0004_django4_update"),
    ]

    operations = [
        migrations.AddField(
            model_name="inactivityping",
            name="last_login_at",
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="leaveofabsence",
            name="created_at",
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="leaveofabsence",
            name="reason",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="Reason for rejecting a request",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="inactivityping",
            name="timestamp",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name="inactivityping",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="leaveofabsence",
            name="notes",
            field=models.TextField(
                blank=True,
                help_text="Description what this leave of absence request is about.",
                verbose_name="description",
            ),
        ),
        migrations.AlterField(
            model_name="leaveofabsence",
            name="user",
            field=models.ForeignKey(
                help_text="The requestor",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="leave_of_absence_requests",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]
